""" Abstract Personality Definition """

import logging
import random
from abc import abstractmethod
from typing import Optional

from pydantic import BaseModel

from rowantree.contracts import Action, ActionQueue, StoreType, UserEvent, UserEventOtherType, UserStore
from rowantree.service.sdk import RowanTreeService

from .abstract_loremaster import AbstractLoremaster


class AbstractPersonality(BaseModel):
    """
    WorldPersonality (Default)
    Generates game world content.

    Attributes
    ----------
    rowantree_service: RowanTreeService
        The Rowan Tree Service Interface.
    loremaster: AbstractLoremaster
        An instance of a story teller for encounter generation.
    """

    rowantree_service: RowanTreeService
    loremaster_service: AbstractLoremaster

    class Config:
        """Pydantic Default Config Over-Rides"""

        arbitrary_types_allowed = True

    ##############
    ## Event Queueing
    ##############

    # TODO: Review the complexity of this
    # pylint: disable=too-many-branches
    def _process_user_event(self, event: Optional[UserEvent], target_user: str) -> None:
        """
        Processes the provided user event (applies the state change of the event)

        Parameters
        ----------
        event: Optional[dict]
            The optional event to process.

        target_user: str
            The target user guid.
        """

        if event is None:
            return

        action_queue: ActionQueue = ActionQueue(queue=[])
        user_stores: dict[StoreType, UserStore] = self.rowantree_service.user_stores_get(user_guid=target_user)

        # process rewards
        rewards_to_remove: list = []
        for reward in event.reward.keys():
            if event.reward[reward] <= 1:
                amount: int = 1
            else:
                amount: int = random.randint(1, event.reward[reward])  # Determine an amount up to the max specified.

            if reward == UserEventOtherType.POPULATION:
                action_queue.queue.append(Action(name="deltaUserPopulationByGUID", arguments=[target_user, amount]))
                event.reward[reward] = amount  # Update to the actual amount
            else:
                if reward in user_stores:
                    store_amt: int = user_stores[reward].amount
                    if store_amt < amount:
                        amount = store_amt

                    action_queue.queue.append(
                        Action(name="deltaUserStoreByStoreNameByGUID", arguments=[target_user, reward, amount])
                    )
                    event.reward[reward] = amount
                else:
                    # Remove skipped types
                    rewards_to_remove.append(reward)
        for reward in rewards_to_remove:
            del event.reward[reward]

        # process curses
        curses_to_remove: list = []
        for curse in event.curse.keys():
            if curse == UserEventOtherType.POPULATION:
                if event.curse[curse] <= 1:
                    pop_amount: int = 1
                else:
                    pop_amount: int = random.randint(1, event.curse[curse])

                user_population: int = self.rowantree_service.user_population_get(user_guid=target_user)
                if user_population < pop_amount:
                    pop_amount: int = user_population

                action_queue.queue.append(
                    Action(name="deltaUserPopulationByGUID", arguments=[target_user, (pop_amount * -1)])
                )
                event.curse[curse] = pop_amount
            else:
                if event.curse[curse] <= 1:
                    amount: int = 1
                else:
                    amount: int = random.randint(1, event.curse[curse])

                if curse in user_stores:
                    store_amt = user_stores[curse].amount
                    if store_amt < amount:
                        amount = store_amt
                else:
                    # Remove skipped types
                    curses_to_remove.append(curse)
                action_queue.queue.append(
                    Action(name="deltaUserStoreByStoreNameByGUID", arguments=[target_user, curse, (amount * -1)])
                )
                event.curse[curse] = amount
        for curse in curses_to_remove:
            del event.curse[curse]

        # Send them the whole event object.
        action_queue.queue.append(
            Action(name="sendUserNotificationByGUID", arguments=[target_user, event.json(by_alias=True)])
        )

        logging.debug(action_queue.json(by_alias=True))
        self.rowantree_service.action_queue_process(queue=action_queue)

    @abstractmethod
    def contemplate(self):
        """Main loop for processing an epoch of events"""



""" Story Teller Definition """
import json
import random
from pathlib import Path
from typing import Any, Optional

from pydantic import ValidationError

from rowantree.contracts import StoreType, UserEvent, UserStore

from ..abstract.abstract_loremaster import AbstractLoremaster


class WorldStoryTeller(AbstractLoremaster):
    """
    The Story Teller
    This class handles world encounters.
    """

    MAX_ENCOUNTER_TRIES = 10

    events: list[UserEvent] = []

    def __init__(self, **data: Any):
        super().__init__(**data)

        with open(file=(Path(__file__).parent / "events.json").resolve(), mode="r", encoding="utf-8") as file:
            global_events_dict = json.load(file)

        for event_dict in global_events_dict["events"]:
            try:
                event: UserEvent = UserEvent.parse_obj(event_dict)
                self.events.append(event)
            except ValidationError as error:
                print(event_dict)
                raise error

    def generate_event(self, user_population: int, user_stores: dict[StoreType, UserStore]) -> Optional[UserEvent]:
        """
        Generates a world event (encounter).

        Parameters
        ----------
        user_population: UserPopulation
            The target user population.
        user_stores: UserStores
            The target user stores.

        Returns
        -------
        outbound_event: Optional[UserEvent]
            An optional encounter for the target user.
        """

        num_events: int = len(self.events)
        requirement_check: bool = False
        counter: int = 0
        new_event: Optional[UserEvent] = None

        while requirement_check is False:
            counter += 1
            event_index = random.randint(1, num_events) - 1
            new_event = self.events[event_index].copy()

            # check requirements
            for requirement in new_event.requirements:
                if requirement == "population":
                    min_required_pop = new_event.requirements[requirement]
                    # logging.debug('reported user population: ' + str(user_population))
                    if user_population >= min_required_pop:
                        requirement_check = True
                else:
                    # assume it is a store - get the current amount of the store for the user
                    min_required_store = new_event.requirements[requirement]
                    if requirement in user_stores:
                        if user_stores[requirement].amount >= min_required_store:
                            requirement_check = True

            # bail out if we've reached the max, no encounters this time
            if counter >= self.MAX_ENCOUNTER_TRIES:
                new_event = None
                requirement_check = True

        if new_event is None:
            return None

        # remove the requirements stanza before we send to over to the client
        new_event.requirements = {}

        return new_event



""" WorldPersonality Definition """

import random
import time
from typing import Optional

from rowantree.contracts import StoreType, UserEvent, UserStore
from rowantree.service.sdk import WorldStatus

from ..abstract.abstract_personality import AbstractPersonality


class WorldPersonality(AbstractPersonality):
    """
    WorldPersonality (Default)
    Generates game world content.

    Attributes
    ----------
    """

    max_sleep_time: int = 3  # in seconds
    encounter_change: int = 100  # in percent

    def contemplate(self) -> None:
        """Reviews active players, and for content."""

        # get active users
        world_status: WorldStatus = self.rowantree_service.world_status_get()
        for target_user in world_status.active_users:
            # Lets add an encounter
            self._encounter(target_user=target_user)

        # now sleep..
        self._slumber()

    def _encounter(self, target_user: str) -> None:
        """
        Cause an encounter.

        Parameters
        ----------
        target_user: str
            The target active user.
        """

        if WorldPersonality._luck(odds=self.encounter_change) is True:
            user_stores: dict[StoreType, UserStore] = self.rowantree_service.user_stores_get(user_guid=target_user)
            user_population: int = self.rowantree_service.user_population_get(user_guid=target_user)

            event: Optional[UserEvent] = self.loremaster_service.generate_event(
                user_population=user_population, user_stores=user_stores
            )
            self._process_user_event(event=event, target_user=target_user)

    def _slumber(self) -> None:
        """
        Sleep wrapper.
        Sleeps for a random amount of time up to the max.
        """

        time.sleep(random.randint(1, self.max_sleep_time))

    @staticmethod
    def _luck(odds: int) -> bool:
        """
        Probability Sample (should something happen)

        Parameters
        ----------
        odds: int
            Change in percent (up to 100%)

        Returns
        -------
        trigger: bool
            Result of the probability evaluation.

        """
        # Ask only for what you truely need and beware you may be granted your wish.
        flip: int = random.randint(1, 100)
        if flip <= odds:
            return True
        return False



""" Abstract Lore Master Definition """
from abc import abstractmethod

from pydantic import BaseModel


class AbstractLoremaster(BaseModel):
    """Abstract Lore Master"""

    @abstractmethod
    def generate_event(self):
        """Generate Event"""
